/*==========================================================================
STDOWIN.C - Re-directs DOS STDIO to specific screen area by intercepting
BIOS INT 10h, func 0Eh.  DOS STDIO routines appear to use this interrupt
exclusively, so any other program that uses STDIO will be constrained to
the window.

NOTES: This routine must do its own interpretation of control characters CR,
LF, TAB.  Use of unsigned char types requires some type casting, but
provides for compatibility with processor register variables.  Microsoft C
uses a row,col screen coordinate convention in text modes with 1,1 being top
left of screen.  The initial arguments to setdoswindow() are passed using
this same 1,1 reference and then converted to system-compatible 0,0.

USES: Primarily as a "viewport" to monitor and interact with dos operations
from shell programs.  I had once seen a shareware program that displayed a
dos "window" while shelled out, which I thought was pretty slick and as it
turned out, also a bit mystifying.  Thorough inspection of several BBS's
programming libraries yielded NOTHING CLOSE.  I decided right then that I
couldn't be the only amateur ever faced with this and that if I solved it,
I'd turn this useful tidbit of knowledge loose in cyberspace.

Some persistent hacking provided the secret - Dos and its companion
externals all use the same bios interrupt for screen output, as does the C
family of STDIO functions.  Turned out to be INT10h, function Eh, (TTY
mimic), which locates screen output based on CURSOR POSITION.  From there
it's simple - hook INT10h and trap anything that might move the cursor out
of bounds.

I found that there is at least one other way to accomplish this constraint
of dos STDIO screen output (see companion file SCRNWIN.C).  It is simple
and it works, although it only constrains the output to a VERTICAL area.  It
does however, have the advantage of handling paged output correctly.

MS QUICK C COMPILING NOTE:  If you're using MS Quick C, this code will NOT
WORK with OPTIMIZATION ON.  For some reason, QC optimizing routines don't
like interrupt functions.  If used as part of a larger program, I suggest
compiling this code in a separate module (with optimization OFF) so that the
other modules can be optimized as desired. In any case, given the liberal
use of assembly in the code, QC's optimization would probably provide little
improvement, (even if it DIDN'T bugger the executable).

This source code is hereby declared PUBLIC DOMAIN.
===========================================================================*/

#include <dos.h>

#include <process.h>    /*for main / demo routine*/

#define TAB_WIDTH (unsigned char)5

void (_interrupt _far *oldint10)(void);     /*pointer to system INT 10h*/

/*prototype replacement INT 10h*/
void _interrupt _far newint10( unsigned _es, unsigned _ds, unsigned _di,
							   unsigned _si, unsigned _bp, unsigned _sp,
							   unsigned _bx, unsigned _dx, unsigned _cx,
							   unsigned _ax, unsigned _ip, unsigned _cs,
							   unsigned _flags );

struct doswin_t     /*all functions reference this window control structure*/
	{
	unsigned char wattrib;  /*color attribute to use in window*/
	unsigned char wtop;     /*top of window*/
	unsigned char wbot;     /*bottom of window*/
	unsigned char wleft;    /*left*/
	unsigned char wright;   /*right*/
	unsigned char wcrow;    /*cursor row*/
	unsigned char wccol;    /*cursor column*/
	}dwindow;

void set_cursor(void);  /*sets curs position from dwindow struct*/
void get_cursor(void);  /*reads curs position into dwindow struct*/

void do_scroll(unsigned char lines);
/*do_scroll takes arguments because it's used both for scrolling 1 line
  at a time, and for multiple lines when initially clearing the window.*/

void do_CR(void);   /*control character handling*/
void do_LF(void);   /*control character handling*/
void do_TAB(void);  /*control character handling*/

/*initializes dwindow structure and sets up INT10h hook*/
void setdoswindow(unsigned char top, unsigned char bot, unsigned char left,
				  unsigned char right, unsigned char attrib);

void canceldoswindow(void); /*restore system default INT10h */

void main()     /*demonstration program*/
{
setdoswindow(11,18,17,65,0x17);

system("command");

canceldoswindow();  /*DON'T forget this step or system will hang*/
}

/*replacement INT function - handle func 0Eh*/
void _interrupt _far newint10( unsigned _es, unsigned _ds, unsigned _di,
							   unsigned _si, unsigned _bp, unsigned _sp,
							   unsigned _bx, unsigned _dx, unsigned _cx,
							   unsigned _ax, unsigned _ip, unsigned _cs,
							   unsigned _flags )
	{
	unsigned char ascii;

	if( (_ax>>8)==0x06 )  /*if scroll*/
		{
		_bx= ((unsigned int)(dwindow.wattrib<<8));
		_cx=( (unsigned int)(dwindow.wtop<<8) + (unsigned int)dwindow.wleft );
		_dx=( (unsigned int)(dwindow.wbot<<8) + (unsigned int)dwindow.wright );
		_chain_intr(oldint10);  /*hand-off to old system INT10*/
		}

	if( (_ax>>8)!=0x0e )        /*check for func 0Eh (TTY output) in ah*/
		_chain_intr(oldint10);  /*if not, hand-off to old system INT*/

	ascii=(unsigned char)(_ax&0x0ff);   /*get character from al*/
	get_cursor();

	switch(ascii)
		{
		case '\x0d':    /*CR*/
		do_CR();
		break;

		case '\x0a':    /*LF*/
		do_LF();
		break;

		case '\x09':    /*TAB*/
		do_TAB();
		break;

		case '\x08':    /*BS*/
		if(dwindow.wccol > dwindow.wleft)
			_chain_intr(oldint10);
		break;

		default:
		if(dwindow.wccol >= dwindow.wright) /*if cursor at right boundary*/
			{
			do_CR();    /*line wrap*/
			do_LF();    /*scroll 1 line*/
			}
		_chain_intr(oldint10);  /*let default INT10h do its thing*/
		break;
		}
	}

void setdoswindow(unsigned char top, unsigned char bot, unsigned char left,
				  unsigned char right, unsigned char attrib)
	{
	/*Initialize dwindow structure to args*/
	/*Also adjust to system (0,0) screen coordinate system*/
	dwindow.wtop=top-(unsigned char)1;
	dwindow.wbot=bot-(unsigned char)1;
	dwindow.wleft=left-(unsigned char)1;
	dwindow.wright=right-(unsigned char)1;

	dwindow.wattrib=attrib;

	_disable();
	oldint10=_dos_getvect(0x10);    /*get system INT address*/
	_dos_setvect(0x10,newint10);    /*install replacement INT*/
	_enable();

	/*clear window*/
	do_scroll( (unsigned char)((dwindow.wbot - dwindow.wtop) + (unsigned char)1) );
	dwindow.wcrow=dwindow.wtop;  /*init cursor to top left*/
	dwindow.wccol=dwindow.wleft;
	set_cursor();
	}

void canceldoswindow(void)
	{
	_disable();
	_dos_setvect(0x10,oldint10);    /*restore system INT*/
	_enable();
	}

void set_cursor(void)
{
_asm
	{
	mov ah, 2
	mov bh, 0
	mov dh, dwindow.wcrow
	mov dl, dwindow.wccol
	int 0x10
	}
}

void do_CR(void)
	{
	dwindow.wccol=dwindow.wleft;
	set_cursor();
	}

void do_LF(void)
	{
	if(dwindow.wcrow<dwindow.wtop)
	   dwindow.wcrow=dwindow.wtop;
	if(dwindow.wcrow==dwindow.wbot)
		do_scroll((unsigned char)1);
	else
		dwindow.wcrow++;
	set_cursor();
	}

void do_TAB(void)
	{
	if( (dwindow.wccol+TAB_WIDTH) > dwindow.wright )
		{
		do_CR();
		do_LF();
		}
	dwindow.wccol=dwindow.wccol+TAB_WIDTH;
	set_cursor();
	}

void do_scroll(unsigned char lines)
	{
	_asm
		{
		mov ah, 6
		mov al, lines
		;(bx,cx,dx will be loaded from within replacement INT10h)
		int 0x10
		}
	}

void get_cursor(void)
    {
    _asm
        {
        mov ah, 3
        mov bh, 0
        int 0x10
        mov dwindow.wcrow, dh
        mov dwindow.wccol, dl
        }
    }
